/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _ASM_X86_HUGETLB_H
#define _ASM_X86_HUGETLB_H

#include <asm/page.h>

#ifndef MODULE
#include <asm/tlbflush.h>

static inline pte_t huge_ptep_clear_flush(struct vm_area_struct *vma,
		unsigned long addr, pte_t *ptep)
{
	struct mm_struct *mm = (vma)->vm_mm;
	pte_t pte;
	pte = ptep_get_and_clear(mm, addr, ptep);
	if (pte_accessible(mm, pte))
		flush_tlb_page(vma, addr);
	return pte;
}
#define __HAVE_ARCH_HUGE_PTEP_CLEAR_FLUSH

static inline void huge_ptep_set_wrprotect(struct mm_struct *mm,
		unsigned long addr, pte_t *ptep)
{
	pax_open_seal();
	ptep_set_wrprotect(mm, addr, ptep);
	pax_close_seal();
}
#define __HAVE_ARCH_HUGE_PTEP_SET_WRPROTECT
#endif

#include <asm-generic/hugetlb.h>

#define hugepages_supported() boot_cpu_has(X86_FEATURE_PSE)

#endif /* _ASM_X86_HUGETLB_H */
