#ifndef AUTOSLAB_H_INCLUDED
#define AUTOSLAB_H_INCLUDED

#include "gcc-common.h"
#include "gcc-ksyms.h"
#include <regex.h>

// usercopy attrs
extern struct attribute_spec alloc_flags_attr;
extern struct attribute_spec alloc_node_attr;
extern struct attribute_spec alloc_attr;

void autoslab_get_alloc_attrs(tree fndecl, tree *count_arg, tree *size_arg, tree *flags_arg, tree *flags_kind, tree *node_arg);

// plugin <-> target interaction
void autoslab_find_decls(void *event_data, void *data __unused);

enum kmalloc_names {
#define KMALLOC_DECL(NAME,NODE,TYPENAME,NOPROF,name,node,typename,noprof) \
	NAME##NODE##TYPENAME##NOPROF,
#include "autoslab.def"
	KMALLOC_NUM
};

enum kmem_call_names {
#define KMEM_CACHE_DECL(NODE,NOPROF,INDEX,TYPE,node,noprof,index,type) \
	KMEM_CACHE_ALLOC##NODE##NOPROF##INDEX##TYPE,
#include "autoslab.def"
	KMEM_CACHE_NUM
};

#ifdef __cplusplus
static inline enum kmem_call_names& operator++(enum kmem_call_names& kmem_call)
{
	if (kmem_call >= KMEM_CACHE_NUM)
		gcc_unreachable();

	kmem_call = static_cast<enum kmem_call_names>(static_cast<unsigned int>(kmem_call) + 1);
	return kmem_call;
}
#endif

extern GTY(()) struct kmem_cache_decl {
	const char * const name;
	tree decl;
} kmem_cache_decls[KMEM_CACHE_NUM];

extern const struct ggc_root_tab gt_ggc_r_autoslab[];

extern const struct kalloc_function {
	const char *kalloc_name;
	enum kmem_call_names kmem_name;
	enum kmem_call_names kmem_name_type;
	enum kmem_call_names kmem_name_index;
	enum kmem_call_names kmem_name_index_type;
} kalloc_functions[KMALLOC_NUM];

// autostack
extern bool autostack;

bool can_convert_to_automatic(gimple_stmt_iterator *gsi,
			      tree size,
			      sbitmap gfp_zones,
			      bool atomic,
			      unsigned HOST_WIDE_INT lhs_align_val);
void autoslab_convert_to_automatic(gimple_stmt_iterator *gsi,
				   tree size,
				   tree flags,
				   sbitmap gfp_zones,
				   unsigned HOST_WIDE_INT lhs_size_val,
				   unsigned HOST_WIDE_INT lhs_align_val);

// autotypename
extern regex_t sizeof_regex;
const char *parse_cache_typename(const char *cache_typename_ptr);
void autoslab_resolve_typenames(void *event_data, void *data __unused);

// misc
extern bool verbose;
tree autoslab_build_var(location_t loc, tree id, tree type, tree ctor);
bool is_cache_alloc(const char *fnname);
tree autoslab_get_alloc_type(tree lhs);

// gcc passes
gimple_opt_pass *make_pass_autoslab(gcc::context *);
simple_ipa_opt_pass *make_pass_autotypename(gcc::context *);

#endif
