/*
 * Copyright 2019-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

void autoslab_get_alloc_attrs(tree fndecl, tree *count_arg, tree *size_arg, tree *flags_arg, tree *flags_kind, tree *node_arg)
{
	tree attr;

	if (count_arg || size_arg) {
		tree count, size;

		attr = lookup_attribute("alloc_size", TYPE_ATTRIBUTES(TREE_TYPE(fndecl)));
		gcc_assert(attr);
		attr = TREE_VALUE(attr);

		count = TREE_VALUE(attr);
		attr = TREE_CHAIN(attr);
		if (attr)
			size = TREE_VALUE(attr);
		else {
			size = count;
			count = NULL_TREE;
		}
		gcc_assert(size);
		if (size_arg)
			*size_arg = size;
		if (count_arg)
			*count_arg = count;
	}

	if (flags_arg || flags_kind) {
		if (flags_arg)
			*flags_arg = NULL_TREE;
		if (flags_kind)
			*flags_kind = NULL_TREE;

		attr = lookup_attribute("alloc_flags", DECL_ATTRIBUTES(fndecl));
		if (attr) {
			if (flags_arg) {
				*flags_arg = TREE_VALUE(TREE_VALUE(attr));
				gcc_assert(*flags_arg);
			}
			if (flags_kind) {
				*flags_kind = TREE_VALUE(TREE_CHAIN(TREE_VALUE(attr)));
				gcc_assert(*flags_kind);
			}
		}
	}

	if (node_arg) {
		*node_arg = NULL_TREE;

		attr = lookup_attribute("alloc_node", DECL_ATTRIBUTES(fndecl));
		if (attr) {
			*node_arg = TREE_VALUE(TREE_VALUE(attr));
			gcc_assert(*node_arg);
		}
	}
}

static tree handle_alloc_flags_attribute(tree *node, tree name, tree args, int flags __unused, bool *no_add_attrs)
{
	location_t loc;

	loc = DECL_SOURCE_LOCATION(*node);
	switch (TREE_CODE(*node)) {
	default:
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute only applies to functions", name);
		break;

	case FUNCTION_DECL:
		break;
	}

	if (TREE_CODE(TREE_VALUE(args)) != INTEGER_CST) {
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute requires a constant for the arg parameter", name);
	}

	if (TREE_CODE(TREE_VALUE(TREE_CHAIN(args))) != STRING_CST) {
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute requires a string for the kind parameter", name);
	}

	return NULL_TREE;
}

static tree handle_alloc_node_attribute(tree *node, tree name, tree args, int flags __unused, bool *no_add_attrs)
{
	location_t loc;

	loc = DECL_SOURCE_LOCATION(*node);
	switch (TREE_CODE(*node)) {
	default:
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute only applies to functions", name);
		break;

	case FUNCTION_DECL:
		break;
	}

	if (TREE_CODE(TREE_VALUE(args)) != INTEGER_CST) {
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute requires a constant for the arg parameter", name);
	}

	return NULL_TREE;
}

static tree handle_alloc_attribute(tree *node, tree name, tree args, int flags __unused, bool *no_add_attrs)
{
	location_t loc;

	loc = DECL_SOURCE_LOCATION(*node);
	switch (TREE_CODE(*node)) {
	default:
		*no_add_attrs = true;
		error_at(loc, "the %qE attribute only applies to functions", name);
		break;

	case FUNCTION_DECL:
		break;
	}

	return NULL_TREE;
}

struct attribute_spec alloc_flags_attr = ATTRIBUTE_SPEC_INIT("alloc_flags", 2, 2, true, false, false, handle_alloc_flags_attribute, false, NULL);
struct attribute_spec alloc_node_attr = ATTRIBUTE_SPEC_INIT("alloc_node", 1, 1, true, false, false, handle_alloc_node_attribute, false, NULL);
struct attribute_spec alloc_attr = ATTRIBUTE_SPEC_INIT("alloc", 0, 0, true, false, false, handle_alloc_attribute, false, NULL);
