/*
 * Copyright 2021-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * gcc does not recognize/handle tail calls to noreturn functions which upsets
 * simpleton tools like objtool, help it by emitting a control flow terminating
 * insn after noreturn calls
 */

#include "utilities.h"

#ifdef TARGET_386
unsigned int fix_tailc_execute(void)
{
	rtx_insn *insn;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		rtx_insn *stuffing;

		if (INSN_DELETED_P(insn))
			continue;

		// rtl match (call_insn (set (reg) (call (mem))))
		if (!CALL_P(insn))
			continue;
		if (SIBLING_CALL_P(insn))
			continue;
		if (!find_reg_note(insn, REG_NORETURN, NULL))
			continue;

#if BUILDING_GCC_VERSION < 8000
		insn = next_nonnote_insn_bb(insn);
#else
		insn = NEXT_INSN(insn);
#endif

		gcc_assert(insn && BARRIER_P(insn));

		start_sequence();
		expand_builtin_trap();
		stuffing = get_insns();
		end_sequence();

		emit_insn_before(stuffing, insn);
	}

	return 0;
}

#define PASS_NAME fix_tailc
#define NO_GATE
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_df_verify
#include "gcc-generate-rtl-pass.h"
#endif
